<?php
namespace OCA\Chores\Controller;

use \DateTimeImmutable;

use OCP\IRequest;
use OCP\IUserManager;
use OCP\AppFramework\Http;
use OCP\AppFramework\Http\DataResponse;
use OCP\AppFramework\Controller;
use OCP\AppFramework\Db\DoesNotExistException;

use OCA\Chores\Db\Team;
use OCA\Chores\Db\TeamMapper;
use OCA\Chores\Db\Chore;
use OCA\Chores\Db\ChoreMapper;
use OCA\Chores\Db\Invite;
use OCA\Chores\Db\InviteMapper;
use OCA\Chores\Db\Member;
use OCA\Chores\Db\MemberMapper;

use OCA\Chores\Service\WorklogService;

class APIController extends Controller {
  private $userId;

  private $choreMapper;
  private $inviteMapper;
  private $memberMapper;
  private $teamMapper;
  private $userManager;
  private $worklogService;

  public function __construct($AppName, IRequest $request,
    TeamMapper $teamMapper, ChoreMapper $choreMapper,
    InviteMapper $inviteMapper,
    MemberMapper $memberMapper, WorklogService $worklogService,
    IUserManager $userManager,
    $UserId){
    parent::__construct($AppName, $request);
    $this->teamMapper = $teamMapper;
    $this->choreMapper = $choreMapper;
    $this->inviteMapper = $inviteMapper;
    $this->memberMapper = $memberMapper;
    $this->userManager = $userManager;
    $this->userId = $UserId;

    $this->worklogService = $worklogService;
  }

  /**
   * @NoAdminRequired
   * @NoCSRFRequired
   */
  public function inviteUserToTeam(int $teamId) {
    try {
      $team = $this->checkAccess($this->userId, $teamId);
      $inviteeUserId = $this->request->post['userId'];
      if (strlen($inviteeUserId) === 0) {
        return new DataResponse(["status" => "bad request"], Http::STATUS_BAD_REQUEST);
      }
      if ($this->memberMapper->teamHasMember($team->id, $inviteeUserId)) {
        return new DataResponse(["code" => "e004", "message" => "Invited user is already a member of that team"], Http::STATUS_CONFLICT);
      }
      $user = $this->userManager->get($inviteeUserId);
      if ($user === null) {
        return new DataResponse(["code" => "e003", "message" => "Invitee does not exist"], Http::STATUS_NOT_FOUND);
      }
      $invite = $this->inviteMapper->createInvite($team->id, $inviteeUserId);
      return new DataResponse($invite, Http::STATUS_OK);
    } catch (DoesNotExistException $e) {
      return new DataResponse([], Http::STATUS_NOT_FOUND);
    }
  }

  /**
   * @NoAdminRequired
   * @NoCSRFRequired
   */
  public function getOwnInvitations() {
    return new DataResponse($this->inviteMapper->findByUser($this->userId), Http::STATUS_OK);
  }

  /**
   * @NoAdminRequired
   * @NoCSRFRequired
   */
  public function acceptInvitation() {
    $teamId = intval($this->request->post["teamId"]);
    //$inviteId = $this->request->post["inviteId"];

    $invite = $this->inviteMapper->findByTeamAndUser($teamId, $this->userId);
    if ($invite === null) {
      return new DataResponse(["code" => "e002", "message" => "invite not found"],
        Http::STATUS_NOT_FOUND);
    }
    $this->memberMapper->addTeamMember($invite->getTeamId(), $this->userId);
    $this->inviteMapper->delete($invite);

    $team = $this->checkAccess($this->userId, $teamId);

    return new DataResponse($team, Http::STATUS_OK);
  }

  /**
   * @NoAdminRequired
   * @NoCSRFRequired
   */
  public function getOwnTeam() {
    return $this->getTeam();
  }

  /**
   * @NoAdminRequired
   * @NoCSRFRequired
   */
  public function getChores(int $teamId) {
    try {
      $team = $this->checkAccess($this->userId, $teamId);
      $chores = $this->choreMapper->findAll($team->id);
      return new DataResponse($chores, Http::STATUS_OK);
    } catch (DoesNotExistException $e) {
      return new DataResponse([], Http::STATUS_NOT_FOUND);;
    }
  }

  private function checkAccess(string $userId, int $teamId): Team {
    $team = $this->teamMapper->find($userId);
    if ($teamId != $team->id) {
      throw new DoesNotExistException("not found");
    }
    return $team;
  }

  /**
   * @NoAdminRequired
   * @NoCSRFRequired
   */
  public function patchChore(int $teamId, int $choreId) {
    try {
      $team = $this->checkAccess($this->userId, $teamId);
      $userId = $this->userId;

      // throws DoesNotExistException if chore does not belong to team:
      $originalChore = $this->choreMapper->findById($team->id, $choreId);

      $updatedChore = (array)($this->request->patch);
      // remove unnecessary request params from JSON body:
      unset($updatedChore["_route"]);
      unset($updatedChore["teamId"]);
      unset($updatedChore["choreId"]);

      $originalChore->updateFromJson($updatedChore);
      $this->choreMapper->update($originalChore);
      return new DataResponse($originalChore, Http::STATUS_OK);
    } catch (DoesNotExistException $e) {
      return new DataResponse([], Http::STATUS_NOT_FOUND);;
    }
  }

  /**
   * @NoAdminRequired
   * @NoCSRFRequired
   */
  public function deleteChore(int $teamId, int $choreId) {
    try {
      $team = $this->checkAccess($this->userId, $teamId);
      $userId = $this->userId;

      // throws DoesNotExistException if chore does not belong to team:
      $chore = $this->choreMapper->findById($team->id, $choreId);

      $this->choreMapper->markDeleted($chore->id);
      return new DataResponse([], Http::STATUS_NO_CONTENT);
    } catch (DoesNotExistException $e) {
      return new DataResponse([], Http::STATUS_NOT_FOUND);;
    }
  }

  /**
   * @NoAdminRequired
   * @NoCSRFRequired
   */
  public function postChores(int $teamId) {
    try {
      $team = $this->checkAccess($this->userId, $teamId);
      $userId = $this->userId;
      $choresPOST = $this->request->getParam("chores");
      $chores = array_map(function($params) use ($team, $userId) {
        return Chore::fromParams([
          "teamId" => $team->id,
          "name" => $params["name"],
          "assigneeUid" => $userId,
          "points" => $params["points"],
          "due" => $params["due"],
          "repeat" => $params["repeat"],
        ]);
      }, $choresPOST);

      $allResponses = [];
      foreach ($chores as $chore) {
        try {
          $this->choreMapper->insert($chore);
          array_push($allResponses, [
            "success" => true,
            "id" => $chore->id,
          ]);
        } catch (\Exception $e) {
          error_log("unable to add chore:" . $e);
          array_push($allResponses, [
            "success" => false,
            "id" => null,
            "error" => "unable to persist chore",
          ]);
        }
      }
      return new DataResponse(["result" => $allResponses], Http::STATUS_OK);
    } catch (DoesNotExistException $e) {
      return new DataResponse([], Http::STATUS_NOT_FOUND);;
    }
  }

  /**
   * @NoAdminRequired
   * @NoCSRFRequired
   */
  public function createTeam() {
    try {
      $team = $this->teamMapper->find($this->userId);
      return new DataResponse([
        "code" => "e001",
        "message" => "User is already member of a team"
      ], Http::STATUS_CONFLICT);
    } catch (DoesNotExistException $e) {
      // User is not yet member of a team, continue
    }
    $team = Team::fromParams([
      "teamName" => $this->request->post["name"],
      "ownerUserId" => $this->userId,
      "content" => "",
    ]);
    $team = $this->teamMapper->insert($team);
    $founderMember = Member::fromParams([
      "teamId" => $team->id,
      "memberUserId" => $this->userId,
      "points" => 0,
    ]);
    $founderMember = $this->memberMapper->insert($founderMember);
    $team->setMembers([$founderMember]);
    return new DataResponse($team, Http::STATUS_CREATED);;
  }

  /**
   * @NoAdminRequired
   * @NoCSRFRequired
   */
  public function getTeam() {
    try {
      $team = $this->teamMapper->find($this->userId);
      $members = $this->memberMapper->findByTeam($team->id);
      foreach ($members as $member) {
        $user = $this->userManager->get($member->getMemberUserId());
        if ($user !== null) {
          $member->setDisplayName($user->getDisplayName());
        }
      }
      $team->setMembers($members);
      $team->setInvites($this->inviteMapper->findByTeam($team->id));
      return new DataResponse($team, Http::STATUS_OK);
    } catch (DoesNotExistException $e) {
      if ($this->request->getParam("no_content") == "404") {
        return new DataResponse([], Http::STATUS_NOT_FOUND);
      } else {
        return new DataResponse([], Http::STATUS_NO_CONTENT);
      }
    }
  }

  /**
   * @NoAdminRequired
   * @NoCSRFRequired
   */
  public function submitWork(string $teamId) {
    $team = $this->checkAccess($this->userId, $teamId);
    $work = $this->request->getParam("work");

    $ok = [];
    $error = [];

    $now = new DateTimeImmutable("now", new \DateTimeZone("UTC"));

    foreach ($work as $workItem) {
      try {
        $this->worklogService->submitWorkAndAddPoints($team->id, $workItem, $now);
        array_push($ok, $workItem["id"]);
      } catch (\Exception $e) {
        array_push($error, $workItem["id"]);
      }
    }

    return new DataResponse([
      "ok" => $ok,
      "error" => $error,
    ], Http::STATUS_OK);

  }

  /**
   * @NoAdminRequired
   * @NoCSRFRequired
   */
  public function getWorklog(string $teamId) {
    $team = $this->checkAccess($this->userId, $teamId);

    $worklog = $this->worklogService->getWorklog($team->id);
    return new DataResponse([
      "worklog" => $worklog,
    ], Http::STATUS_OK);

  }

}
